--- old/usr/src/uts/common/fs/zfs/arc.c	Tue Nov 11 06:35:52 2014
+++ new/usr/src/uts/common/fs/zfs/arc.c	Tue Nov 11 06:35:52 2014
@@ -2519,6 +2519,7 @@
 
 	arc_adapt(size, state);
 
+top:
 	/*
 	 * We have not yet reached cache maximum size,
 	 * just allocate a new buffer.
@@ -2525,15 +2526,24 @@
 	 */
 	if (!arc_evict_needed(type)) {
 		if (type == ARC_BUFC_METADATA) {
-			buf->b_data = zio_buf_alloc(size);
-			arc_space_consume(size, ARC_SPACE_DATA);
+			buf->b_data = zio_buf_alloc_canfail(size);
+			if (buf->b_data != NULL) {
+				arc_space_consume(size, ARC_SPACE_DATA);
+				goto out;
+			}
 		} else {
 			ASSERT(type == ARC_BUFC_DATA);
-			buf->b_data = zio_data_buf_alloc(size);
-			ARCSTAT_INCR(arcstat_data_size, size);
-			atomic_add_64(&arc_size, size);
+			buf->b_data = zio_data_buf_alloc_canfail(size);
+			if (buf->b_data != NULL) {
+				ARCSTAT_INCR(arcstat_data_size, size);
+				atomic_add_64(&arc_size, size);
+				goto out;
+			}
 		}
-		goto out;
+		/*
+		 * Memory allocation failed probably due to excessive
+		 * fragmentation, we need to evict regardless.
+		 */
 	}
 
 	/*
@@ -2556,16 +2566,8 @@
 		    mfu_space > arc_mfu->arcs_size) ? arc_mru : arc_mfu;
 	}
 	if ((buf->b_data = arc_evict(state, NULL, size, TRUE, type)) == NULL) {
-		if (type == ARC_BUFC_METADATA) {
-			buf->b_data = zio_buf_alloc(size);
-			arc_space_consume(size, ARC_SPACE_DATA);
-		} else {
-			ASSERT(type == ARC_BUFC_DATA);
-			buf->b_data = zio_data_buf_alloc(size);
-			ARCSTAT_INCR(arcstat_data_size, size);
-			atomic_add_64(&arc_size, size);
-		}
 		ARCSTAT_BUMP(arcstat_recycle_miss);
+		goto top;
 	}
 	ASSERT(buf->b_data != NULL);
 out:
--- old/usr/src/uts/common/fs/zfs/sys/zio.h	Tue Nov 11 06:35:53 2014
+++ new/usr/src/uts/common/fs/zfs/sys/zio.h	Tue Nov 11 06:35:53 2014
@@ -511,8 +511,10 @@
 extern void zio_add_child(zio_t *pio, zio_t *cio);
 
 extern void *zio_buf_alloc(size_t size);
+extern void *zio_buf_alloc_canfail(size_t size);
 extern void zio_buf_free(void *buf, size_t size);
 extern void *zio_data_buf_alloc(size_t size);
+extern void *zio_data_buf_alloc_canfail(size_t size);
 extern void zio_data_buf_free(void *buf, size_t size);
 
 extern void zio_resubmit_stage_async(void *);
--- old/usr/src/uts/common/fs/zfs/zio.c	Tue Nov 11 06:35:53 2014
+++ new/usr/src/uts/common/fs/zfs/zio.c	Tue Nov 11 06:35:53 2014
@@ -221,6 +221,20 @@
 }
 
 /*
+ * Same as zio_buf_alloc, but won't sleep in case memory cannot be allocated
+ * and will instead return immediately with a failure.
+ */
+void *
+zio_buf_alloc_canfail(size_t size)
+{
+	size_t c = (size - 1) >> SPA_MINBLOCKSHIFT;
+
+	ASSERT(c < SPA_MAXBLOCKSIZE >> SPA_MINBLOCKSHIFT);
+
+	return (kmem_cache_alloc(zio_buf_cache[c], KM_NOSLEEP | KM_NORMALPRI));
+}
+
+/*
  * Use zio_data_buf_alloc to allocate data.  The data will not appear in a
  * crashdump if the kernel panics.  This exists so that we will limit the amount
  * of ZFS data that shows up in a kernel crashdump.  (Thus reducing the amount
@@ -236,6 +250,21 @@
 	return (kmem_cache_alloc(zio_data_buf_cache[c], KM_PUSHPAGE));
 }
 
+/*
+ * Same as zio_data_buf_alloc, but won't sleep in case memory cannot be
+ * allocated and will instead return immediately with a failure.
+ */
+void *
+zio_data_buf_alloc_canfail(size_t size)
+{
+	size_t c = (size - 1) >> SPA_MINBLOCKSHIFT;
+
+	ASSERT(c < SPA_MAXBLOCKSIZE >> SPA_MINBLOCKSHIFT);
+
+	return (kmem_cache_alloc(zio_data_buf_cache[c],
+	    KM_NOSLEEP | KM_NORMALPRI));
+}
+
 void
 zio_buf_free(void *buf, size_t size)
 {
